/*****************************************************************************
 * vim:tw=78
 *
 * Copyright (C) 2021 Bjarne von Horn (vh at igh dot de).
 *
 * This file is part of the PdCom library.
 *
 * The PdCom library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The PdCom library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the PdCom library. If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <fstream>
#include <gnutls/gnutls.h>
#include <iostream>
#include <pdcom5/PosixProcess.h>
#include <pdcom5/SecureProcess.h>
#include <pdcom5/Variable.h>
#include <poll.h>

class MyProcess : public PdCom::SecureProcess, private PdCom::PosixProcess
{
    int secureRead(char *buf, int count) override
    {
        return posixRead(buf, count);
    }
    void secureWrite(const char *buf, size_t count) override
    {
        posixWriteDirect(buf, count);
    }

    void listReply(
            std::vector<PdCom::Variable> vars /* variables */,
            std::vector<std::string> /* directories */) override
    {
        for (const auto &v : vars)
            std::cout << "Found var " << v.getPath() << "\n";
        running_ = false;
    }

    void connected()
    {
        std::cout << "Connected!" << std::endl;
        list("");
    }

  public:
    void waitForSocket()
    {
        fd_set fds;
        FD_ZERO(&fds);
        FD_SET(fd_, &fds);
        select(fd_ + 1, &fds, NULL, NULL, NULL);
    }

    MyProcess(const std::string &ca_pem, const char *host = "localhost") :
        PdCom::SecureProcess({EncryptionDetails::Default, ca_pem, host}),
        PdCom::PosixProcess(host, 4567)
    {
        handshake();
    }

    bool running_ = true;
};

int main(int argc, char *argv[])
{
    if (argc < 3) {
        std::cerr << "Usage: " << argv[0] << " <hostname> <Path to CA.pem>\n";
        return -1;
    }
    gnutls_global_init();
    std::ifstream ca_file(argv[2]);
    std::string ca;
    getline(ca_file, ca, '\0');
    MyProcess p(ca, argv[1]);
    while (p.running_) {
        p.waitForSocket();
        p.asyncData();
    }
    p.bye();
    return 0;
}
