/*****************************************************************************
 *
 * Copyright (C) 2021 Bjarne von Horn (vh at igh dot de).
 *
 * This file is part of the PdCom library.
 *
 * The PdCom library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The PdCom library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the PdCom library. If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif
#include <Python.h>
#include <memory>
#include <pdcom5/Exception.h>
#include <pdcom5/Process.h>
#include <pdcom5/Subscriber.h>
#include <pdcom5/Subscription.h>
#include <pdcom5/Variable.h>
#include <pybind11/chrono.h>
#include <pybind11/numpy.h>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

namespace py = pybind11;

namespace {
PdCom::TypeInfo::DataType convert_dt(py::dtype const np_dt)
{
    using TI = PdCom::TypeInfo::DataType;
    switch (np_dt.kind()) {
        case 'b':
        case 'i':
            switch (np_dt.itemsize()) {
                case 1:
                    return TI::int8_T;
                case 2:
                    return TI::int16_T;
                case 4:
                    return TI::int32_T;
                case 8:
                    return TI::int64_T;
            }
            break;
        case 'B':
        case 'u':
            switch (np_dt.itemsize()) {
                case 1:
                    return TI::uint8_T;
                case 2:
                    return TI::uint16_T;
                case 4:
                    return TI::uint32_T;
                case 8:
                    return TI::uint64_T;
            }
            break;
        case 'f':
            switch (np_dt.itemsize()) {
                case 4:
                    return TI::single_T;
                case 8:
                    return TI::double_T;
            }
    }
    throw std::runtime_error("unknown datatype");
}
py::array create_buf(const PdCom::Variable &v)
{
    using dt = PdCom::TypeInfo::DataType;
    const py::array::ShapeContainer sc {
            static_cast<Py_ssize_t>(v.getSizeInfo().totalElements())};
    switch (v.getTypeInfo().type) {
        // case dt::boolean_T:
        case dt::uint8_T:
            return {py::dtype("B"), sc};
        case dt::int8_T:
            return {py::dtype("b"), sc};
        case dt::uint16_T:
            return {py::dtype("u2"), sc};
        case dt::int16_T:
            return {py::dtype("i2"), sc};
        case dt::uint32_T:
            return {py::dtype("u4"), sc};
        case dt::int32_T:
            return {py::dtype("i4"), sc};
        case dt::uint64_T:
            return {py::dtype("u8"), sc};
        case dt::int64_T:
            return {py::dtype("i8"), sc};
        case dt::double_T:
            return {py::dtype("f8"), sc};
        case dt::single_T:
            return {py::dtype("f4"), sc};
        default:
            throw std::runtime_error("unknown type id");
    }
}
}  // namespace

class WrappedProcess;

class WrappedVariable : public PdCom::Variable
{
    std::shared_ptr<WrappedProcess> process_;

  public:
    std::shared_ptr<WrappedProcess> getProcess() const { return process_; }

    WrappedVariable(
            PdCom::Variable v,
            std::shared_ptr<WrappedProcess> process) :
        PdCom::Variable(std::move(v)), process_(std::move(process))
    {}

    bool setValue(py::array values, size_t offset)
    {
        if (values.size() < 0 or offset >= static_cast<size_t>(values.size()))
            throw std::range_error("Offset out of Range");
        // TODO(vh-igh) strides!?
        const auto np_dt = values.dtype();
        if (np_dt.has_fields())
            return false;
        const auto pd_type = convert_dt(np_dt);
        return PdCom::Variable::setValue(
                values.data(), pd_type, offset, values.size() - offset);
    }
    py::object getSizeInfoPy()
    {
        py::list ans;
        for (const auto s : getSizeInfo())
            ans.append(s);
        return ans;
    }
};

class PythonSubscriber :

    public std::enable_shared_from_this<PythonSubscriber>,
    public PdCom::Subscriber
{
  public:
    PythonSubscriber(const PdCom::Transmission &tr) :
        std::enable_shared_from_this<PythonSubscriber>(), PdCom::Subscriber(tr)
    {}
};

class SubscriberTrampoline : public PythonSubscriber
{
  public:
    static std::shared_ptr<SubscriberTrampoline>
    create(const PdCom::Transmission &tr)
    {
        return std::make_shared<SubscriberTrampoline>(tr);
    }

    using PythonSubscriber::PythonSubscriber;

    void newValues(std::chrono::nanoseconds time_ns) override
    {
        PYBIND11_OVERRIDE_PURE(void, PythonSubscriber, newValues, time_ns);
    }

    void stateChanged(PdCom::Subscription const &) override;
};


class WrappedProcess :
    public PdCom::Process,
    public std::enable_shared_from_this<WrappedProcess>
{
    using PdCom::Process::Process;

  public:
    static std::shared_ptr<WrappedProcess> create()
    {
        return std::make_shared<WrappedProcess>();
    }

    std::string applicationName() const override
    {
        PYBIND11_OVERRIDE(std::string, PdCom::Process, applicationName, );
    }
    int read(char *buf, int count) override
    {
        PYBIND11_OVERRIDE_PURE(
                int, PdCom::Process, read,
                py::memoryview::from_memory(buf, count, false));
    }
    void write(const char *buf, size_t count) override
    {
        PYBIND11_OVERRIDE_PURE(
                void, PdCom::Process, write,
                py::memoryview::from_memory(buf, count));
    }
    void flush() override
    {
        py::gil_scoped_acquire gil;
        py::function override = py::get_override(
                static_cast<const PdCom::Process *>(this), "flush");
        if (override) {
            override();
        }
    }
    void connected() override
    {
        PYBIND11_OVERRIDE_PURE(void, PdCom::Process, connected, );
    }
    void listReply(
            std::vector<PdCom::Variable> variables,
            std::vector<std::string> directories) override
    {
        py::gil_scoped_acquire gil;
        py::function override = py::get_override(
                static_cast<const PdCom::Process *>(this), "listReply");
        if (override) {
            py::list vars;
            for (auto &var : variables) {
                vars.append(
                        WrappedVariable(std::move(var), shared_from_this()));
            }
            override(vars, directories);
        }
    }
    void findReply(const PdCom::Variable &variable) override
    {
        py::gil_scoped_acquire gil;
        py::function override = py::get_override(
                static_cast<const PdCom::Process *>(this), "findReply");
        if (override) {
            if (variable.empty())
                override(py::none());
            else
                override(std::unique_ptr<WrappedVariable>(
                        new WrappedVariable(variable, shared_from_this())));
        }
    }
    void pingReply() override
    {
        py::gil_scoped_acquire gil;
        py::function override = py::get_override(
                static_cast<const PdCom::Process *>(this), "pingReply");
        if (override) {
            override();
        }
    }
    bool alive() override
    {
        py::gil_scoped_acquire gil;
        py::function override = py::get_override(
                static_cast<const PdCom::Process *>(this), "alive");
        if (override) {
            return py::cast<bool>(override());
        }
        return true;
    }
    void transmitSemaphore(bool state) override
    {
        py::gil_scoped_acquire gil;
        py::function override = py::get_override(
                static_cast<const PdCom::Process *>(this), "transmitSemaphore");
        if (override) {
            override(state);
        }
    }

    void broadcastReply(
            const std::string &message,
            const std::string &attr,
            std::chrono::nanoseconds time_ns,
            const std::string &user) override
    {
        py::gil_scoped_acquire gil;
        py::function override = py::get_override(
                static_cast<const PdCom::Process *>(this), "broadcastReply");
        if (override) {
            override(message, attr, time_ns, user);
        }
    }
};


class WrappedSubscription :
    public std::enable_shared_from_this<WrappedSubscription>,
    public PdCom::Subscription
{
    std::shared_ptr<WrappedProcess> process_;
    std::shared_ptr<PythonSubscriber> mailbox_;

  public:
    WrappedSubscription() = default;
    WrappedSubscription(
            std::shared_ptr<PythonSubscriber> sub,
            const WrappedVariable &var) :
        std::enable_shared_from_this<WrappedSubscription>(),
        PdCom::Subscription(*sub, var),
        process_(var.getProcess()),
        mailbox_(sub)
    {}
    WrappedSubscription(
            std::shared_ptr<PythonSubscriber> sub,
            std::shared_ptr<WrappedProcess> process,
            const std::string &path) :
        std::enable_shared_from_this<WrappedSubscription>(),
        PdCom::Subscription(*sub, *process, path),
        process_(process),
        mailbox_(sub)
    {}

    static std::shared_ptr<WrappedSubscription> create1()
    {
        return std::make_shared<WrappedSubscription>();
    }
    static std::shared_ptr<WrappedSubscription>
    create2(std::shared_ptr<PythonSubscriber> sub, const WrappedVariable &var)
    {
        return std::make_shared<WrappedSubscription>(sub, var);
    }
    static std::shared_ptr<WrappedSubscription>
    create3(std::shared_ptr<PythonSubscriber> sub,
            std::shared_ptr<WrappedProcess> process,
            const std::string &path)
    {
        return std::make_shared<WrappedSubscription>(sub, process, path);
    }


    void cancel()
    {
        static_cast<PdCom::Subscription &>(*this) = PdCom::Subscription();
    }
    WrappedSubscription(WrappedSubscription const &) = delete;
    WrappedSubscription(WrappedSubscription &&)      = default;
    WrappedSubscription &operator=(WrappedSubscription const &) = delete;
    WrappedSubscription &operator=(WrappedSubscription &&) = default;

    virtual ~WrappedSubscription()
    {
        // unsubscribe before Process goes away
        if (getState() != State::Invalid)
            cancel();
    }

    WrappedVariable getVariablePy() { return {getVariable(), process_}; }

    py::object getValuePy()
    {
        auto ans = create_buf(getVariable());
        std::memcpy(
                ans.mutable_data(), getData(),
                getVariable().getSizeInfo().totalElements()
                        * getVariable().getTypeInfo().element_size);
        return ans;
    }
};

void SubscriberTrampoline::stateChanged(PdCom::Subscription const &s)
{
    PYBIND11_OVERRIDE_PURE(
            void, PythonSubscriber, stateChanged,
            const_cast<WrappedSubscription &>(
                    static_cast<WrappedSubscription const &>(s))
                    .shared_from_this());
}

PYBIND11_MODULE(PdComWrapper, m)
{
    using PdCom::Process;
    using PdCom::Subscription;
    using PdCom::Transmission;
    using PdCom::Variable;

    py::class_<Process, std::shared_ptr<Process>, WrappedProcess>(m, "Process")
            .def(py::init<>(&WrappedProcess::create))
            .def_property_readonly("name", &Process::name)
            .def_property_readonly("version", &Process::version)
            .def("asyncData", &Process::asyncData)
            .def("callPendingCallbacks", &Process::callPendingCallbacks)
            .def("list", &Process::list, py::arg("path") = "")
            .def("find", &Process::find, py::arg("path"))
            .def("ping", &Process::ping)
            .def("broadcast", &Process::broadcast, py::arg("message"),
                 py::arg("attr") = "text")
            //
            ;
    py::class_<PdCom::Variable>(m, "_Variable");
    py::class_<WrappedVariable, PdCom::Variable>(m, "Variable")
            .def_property_readonly("empty", &Variable::empty)
            .def_property_readonly("path", &Variable::getPath)
            .def_property_readonly("name", &Variable::getName)
            .def_property_readonly("size_info", &WrappedVariable::getSizeInfo)
            .def("setValue", &WrappedVariable::setValue, py::arg("value"),
                 py::arg("offset") = 0);
    py::class_<WrappedSubscription, std::shared_ptr<WrappedSubscription>> sub(
            m, "Subscription");
    sub.def(py::init<>(&WrappedSubscription::create1))
            .def(py::init<>(&WrappedSubscription::create2),
                 py::arg("subscriber"), py::arg("variable"))
            .def(py::init<>(&WrappedSubscription::create3),
                 py::arg("subscriber"), py::arg("process"), py::arg("path"))
            .def("cancel", &WrappedSubscription::cancel)
            .def_property_readonly(
                    "variable", &WrappedSubscription::getVariablePy)
            .def_property_readonly("value", &WrappedSubscription::getValuePy)
            .def_property_readonly("state", [](const WrappedSubscription &s) {
                return s.getState();
            });
    py::class_<
            PythonSubscriber, std::shared_ptr<PythonSubscriber>,
            SubscriberTrampoline>(m, "Subscriber")
            .def(py::init<>(&SubscriberTrampoline::create),
                 py::arg("transmission"))
            .def_property_readonly(
                    "transmission", &PdCom::Subscriber::getTransmission);

    py::enum_<Subscription::State>(sub, "State")
            .value("Pending", Subscription::State::Pending)
            .value("Active", Subscription::State::Active)
            .value("Invalid", Subscription::State::Invalid);
    py::class_<Transmission>(m, "Transmission")
            .def(py::init<std::chrono::duration<double>>(), py::arg("duration"))
            .def_property_readonly_static(
                    "poll_mode",
                    [](py::object) { return Transmission {PdCom::poll_mode}; })
            .def_property_readonly_static("event_mode", [](py::object) {
                return Transmission {PdCom::event_mode};
            });

    const py::handle ex_base =
            py::register_exception<PdCom::Exception>(m, "Exception");
    py::register_exception<PdCom::InternalError>(m, "InternalError", ex_base);
    py::register_exception<PdCom::InvalidArgument>(
            m, "InvalidArgument", ex_base);
    py::register_exception<PdCom::EmptyVariable>(m, "EmptyVariable", ex_base);
    py::register_exception<PdCom::InvalidSubscription>(
            m, "InvalidSubscription", ex_base);
    py::register_exception<PdCom::NotConnected>(m, "NotConnected", ex_base);
    py::register_exception<PdCom::ProcessGoneAway>(
            m, "ProcessGoneAway", ex_base);
    py::register_exception<PdCom::ProtocolError>(m, "ProtocolError", ex_base);
}
