/*****************************************************************************
 *
 * Copyright (C) 2015-2016  Richard Hacker (lerichi at gmx dot net)
 *
 * This file is part of the PdCom library.
 *
 * The PdCom library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The PdCom library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the PdCom library. If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include "ProtocolHandler.h"

#include "../Debug.h"
#include "../SpyLayer.h"
#include "Channel.h"
#include "DirNode.h"
#include "Parameter.h"
#include "Subscription.h"
#include "expat_wrapper_impl.h"

#include <algorithm>
#include <cstring>
#include <errno.h>
#include <mutex>
#include <ostream>
#include <pdcom5/Exception.h>
#include <pdcom5/Subscriber.h>
#include <sstream>
#include <stdexcept>
#include <typeinfo>

#ifndef _WIN32

#include <pthread.h>
namespace {
class PthreadCancelDisabler
{
    int cancel_state_ = 0;

  public:
    PthreadCancelDisabler()
    {
        pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel_state_);
    }
    ~PthreadCancelDisabler() { pthread_setcancelstate(cancel_state_, nullptr); }
};
}  // namespace

#else  // _WIN32
namespace {
struct PthreadCancelDisabler
{};
}  // namespace

#endif  // _WIN32

#ifdef PDC_DEBUG
#include <stdlib.h>  // getenv(), atoi()
#endif

using PdCom::impl::MsrProto::Channel;
using PdCom::impl::MsrProto::ExpatWrapper;
using PdCom::impl::MsrProto::Parameter;
using PdCom::impl::MsrProto::ProtocolHandler;
using PdCom::impl::MsrProto::Variable;
using PdCom::impl::MsrProto::XmlStream;

template <typename T>
static void unwind_safe_execute(T &&cmd)
{
#if __cpp_lib_uncaught_exceptions
    const int pending_exceptions = std::uncaught_exceptions();
#else
    const bool pending_exception = std::uncaught_exception();
#endif
    try {
        cmd();
    }
    catch (...) {
        // rethrow exception if no stack unwinding is in progress
#if __cpp_lib_uncaught_exceptions
        if (pending_exceptions == std::uncaught_exceptions())
#else
        if (!pending_exception and !std::uncaught_exception())
#endif
            throw;
    }
}

namespace {
struct Attribute
{ /*{{{*/
    Attribute(const char **atts) : atts(atts) {}

    const char *find(const char *arg)
    {
        for (size_t i = 0; arg and atts[i]; i += 2) {
            if (!strcmp(atts[i], arg))
                return atts[i + 1];
        }

        return nullStr;
    }

    bool getString(const char *arg, const char *&val)
    {
        const char *attr = find(arg);

        if (attr == nullStr)
            return false;

        val = attr;
        return true;
    };

    bool getUInt(const char *arg, unsigned int &val)
    {
        const char *attr = find(arg);

        // Only test for numbers, even leading '+' is ignored
        if (*attr < '0' or *attr > '9')
            return false;

        val = atol(attr);
        return true;
    }

    bool getInt(const char *arg, int &val)
    {
        const char *attr = find(arg);

        // Only test for numbers, even leading '+' is ignored
        if ((*attr < '0' or *attr > '9') and *attr != '-')
            return false;

        val = atoi(attr);
        return true;
    }

    bool isTrue(const char *arg)
    {
        unsigned int val;
        return getUInt(arg, val) and val;
    }

    bool getDouble(const char *arg, double &val)
    {
        const char *attr = find(arg);

        if (*attr)
            val = atof(attr);

        return *attr;
    }

    std::chrono::nanoseconds getTime_ns(const char *arg)
    {
        const char *attr = find(arg);

        int64_t time_ns = 0;
        size_t i        = 0;
        do {
            if (!*attr or (!i and *attr == '.')) {
                if (!i)
                    i = 10;
            }
            else if (*attr < '0' or *attr > '9')
                return std::chrono::nanoseconds {0};

            if (*attr != '.') {
                if (*attr or i != 10)
                    time_ns *= 10;

                if (*attr)
                    time_ns += *attr - '0';
            }

            if (*attr)
                ++attr;

        } while (*attr == '.' or !i or --i);

        return std::chrono::nanoseconds {time_ns};
    }

    static const char *const nullStr;

    const char **atts;
}; /*}}}*/
}  // namespace

const char *const Attribute::nullStr = "";

XmlStream::XmlStream(PdCom::impl::Process *process, std::streambuf *sb) :
    std::ostream(sb), process(process)
{
    exceptions(std::ios_base::badbit);
}
void XmlStream::lock()
{
    process->transmitSemaphore(true);
}
void XmlStream::unlock()
{
    process->transmitSemaphore(false);
}


namespace {
struct XmlCommand
{ /*{{{*/
    struct Attribute
    {
        // Create an attribute for a command.
        Attribute(XmlCommand &command, const char *name) : cout(command.cout)
        {
            // send the preamble
            cout << ' ' << name << "=\"";
        }
        ~Attribute() noexcept(false)
        {
            unwind_safe_execute([this] { cout << '"'; });
        }

        std::ostream &cout;
    };

    XmlCommand(XmlStream &cout, const char *command) :
        cout(cout), guard(cout), p_flush(true)
    {
        cout << '<' << command;
    }

    ~XmlCommand() noexcept(false)
    {
        unwind_safe_execute([this] {
            cout << ">\r\n";
            if (p_flush)
                cout << std::flush;
        });
    }

    XmlCommand &addEscapedAttribute(const char *name, const std::string &str)
    {
        cout << ' ' << name << '=' << '"';

        for (const char c : str) {
            switch (c) {
                case '"':
                    cout << "&quot;";
                    break;
                case '<':
                    cout << "&lt;";
                    break;
                case '&':
                    cout << "&amp;";
                    break;
                case '\0':
                    throw PdCom::InvalidArgument("Value must not contain NULL");
                default:
                    cout << c;
            }
        }

        cout << '"';

        return *this;
    }

    template <typename T>
    XmlCommand &addAttribute(const char *name, T const &val, bool active = true)
    {
        if (active)
            cout << ' ' << name << "=\"" << val << '"';
        return *this;
    }

    template <typename T>
    XmlCommand &addAttributeStream(const char *name, T &&callback)
    {
        cout << ' ' << name << "=\"";
        callback(cout);
        cout << '"';
        return *this;
    }

    XmlCommand &noFlush()
    {
        p_flush = false;
        return *this;
    }

    XmlCommand &setId(const char *id)
    {
        if (id and *id)
            cout << " id=\"" << id << '"';

        return *this;
    }

    XmlStream &cout;
    std::lock_guard<XmlStream> guard;
    bool p_flush;
}; /*}}}*/
}  // namespace

///////////////////////////////////////////////////////////////////////////
ProtocolHandler::ProtocolHandler(PdCom::impl::Process *process, IOLayer *io) :
    PdCom::impl::ProtocolHandler(process),
    StreambufLayer(io),
    ExpatWrapper<ProtocolHandler>(),
    cout(process, this),
    root(std::make_shared<DirNode>())
{ /*{{{*/

#if PDC_DEBUG
    {
        char *spy = getenv("SPY");
        if (spy and atoi(spy))
            insert(new SpyLayer(io));
    }
#endif

    if (!parse("<xml>", 5))
        throw Exception("setting up xml parser failed");
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
ProtocolHandler::~ProtocolHandler()
{ /*{{{*/
    log_debug("killed %p", static_cast<void *>(this));
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::xmlError(const char *err_msg)
{ /*{{{*/
    if (state != StartUp and state != WaitForConnected)
        throw ProtocolError(std::string("Fatal XML parsing error: ") + err_msg);
} /*}}}*/

#if PDC_DEBUG
const char *statename[] = {
        "StartUp",
        "Idle",
        "GetListing",
        "WaitForConnected",
};
#endif

///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::startElement(const char *name, const char **atts)
{ /*{{{*/
    const char *id;

    switch (state) {
        case StartUp:
            if (!strcmp(name, "xml") and level() == 1)
                state = WaitForConnected;
            break;

        case WaitForConnected:
            if (!strcmp(name, "connected") and level() == 2) {
                Attribute a(atts);
                feature   = a.find("features");
                m_name    = a.find("app");
                m_version = a.find("appversion");

                // if login is mandatory, <saslauth> is sent automatically by
                // server
                if (feature.login and !strcmp(a.find("login"), "mandatory")) {
                    if (process->auth_manager_) {
                        state = WaitForLogin;
                        break;
                    }
                    else
                        throw LoginRequired();
                }
                XmlCommand(cout, "remote_host")
                        .addAttribute("access", "allow")
                        .addEscapedAttribute(
                                "applicationname", process->applicationName())
                        .addEscapedAttribute("hostname", process->hostname())
                        .setId("init");

                state = Idle;
            }
            break;

        case WaitForLogin:
            if (level() == 2 and !strcmp(name, "saslauth")
                and processLogin(atts)) {
                XmlCommand(cout, "remote_host")
                        .addAttribute("access", "allow")
                        .addEscapedAttribute(
                                "applicationname", process->applicationName())
                        .addEscapedAttribute("hostname", process->hostname())
                        .setId("init");

                state = Idle;
            }
            break;

        case Idle:
            if (level() != 2) {
                log_debug("Unknown XML tag in Idle: %s", name);
                break;
            }

            // find() returns "" if id is not found
            id = Attribute(atts).find("id");

            if (!strcmp(name, "data")) {
                Attribute a(atts);

                unsigned int dataGroupId = 0;
                a.getUInt("group", dataGroupId);

                dataTime = a.getTime_ns("time");

                if (!dataGroupId) {
                    // Attribute "group" not found. Event transmission
                    state = ReadEvent;
                }
                else {
                    periodic_subscriptions_.startNewDataRecieving(
                            dataGroupId, variable_cache_.lock());
                    state = ReadData;
                }
            }
            else if (!strcmp(name, "pu")) {
                unsigned int index;

                if (!feature.xsap and Attribute(atts).getUInt("index", index)) {
                    if (parameter_event_.find(index)
                        != parameter_event_.end()) {
                        pollParameter(index, "pollParameterEvent");
                    }
                }
            }
            else if (!strcmp(name, "parameter")) {
                Parameter *p = getParameter(atts);
                if (p) {
                    Attribute a(atts);
                    const char *const hexvalue = a.find("hexvalue");
                    const auto time_ns         = a.getTime_ns("time");
                    unsigned int pm;
                    if (!strcmp(id, "pollParameter")
                        and !pending_parameter_polls_.empty()) {
                        auto cache = variable_cache_.lock();
                        cache[0].readFromHexDec(hexvalue, *p);
                        auto sub_collection =
                                std::move(pending_parameter_polls_.front());
                        pending_parameter_polls_.pop_front();
                        sub_collection.subscriptions_.readData(
                                cache[0].data(), time_ns);
                    }
                    else if (
                            !feature.xsap
                            and !strcmp(id, "pollParameterEvent")) {
                        auto cache = variable_cache_.lock();
                        cache[0].readFromHexDec(hexvalue, *p);
                        parameter_event_[p->index_].readData(
                                cache[0].data(), time_ns);
                    }
                    else if (!strcmp(id, "pendingParameter"))
                        pendingAck(p);
                    else if (a.getUInt("pm", pm) && pm == 1) {
                        auto cache = variable_cache_.lock();
                        cache[0].readFromHexDec(hexvalue, *p);
                        parameter_event_[p->index_].readData(
                                cache[0].data(), time_ns);
                    }
                }
            }
            else if (!strcmp(name, "channel")) {
                Channel *c = getChannel(atts);
                if (c) {
                    if (!strcmp(id, "pollChannel")
                        and !pending_channel_polls_.empty()) {
                        // can be nullptr if subscription was deleted in the
                        // meantime
                        auto channel_poll =
                                std::move(pending_channel_polls_.front());
                        pending_channel_polls_.pop_front();
                        Attribute a(atts);
                        auto cache = variable_cache_.lock();
                        cache[0].readFromString(a.find("value"), *c);
                        for (const auto &weak_sub : channel_poll.subscriptions_)
                            if (const auto sub = weak_sub.lock()) {
                                sub->readData(cache[0].data());
                                sub->newValues(a.getTime_ns("time"));
                            }
                    }
                    else if (!strcmp(id, "pendingChannel"))
                        pendingAck(c);
                }
            }
            else if (!strcmp(name, "ack")) {
                // Check which ack was called
                if (!strcmp(id, "xsadQ"))
                    xsadAck();
                else if (!strcmp(id, "xsapQ"))
                    xsapAck();
                else if (!strcmp(id, "pendingChannel"))
                    pendingAck(nullptr);
                else if (!strcmp(id, "findQ"))
                    processFindRequest();
                else if (!strcmp(id, "listQ")) {
                    fullListing = Cached;

                    // Process all pending list and find requests
                    while (processListRequest())
                        ;
                    while (processFindRequest())
                        ;
                }
                else if (!strcmp(id, "init"))
                    process->connected();
            }
            else if (
                    !strcmp(name, "listing") or !strcmp(name, "channels")
                    or !strcmp(name, "parameters")) {
                state = GetListing;
            }
            else if (!strcmp(name, "saslauth"))
                processLogin(atts);
            else if (!strcmp(name, "ping"))
                process->pingReply();
            else if (!strcmp(name, "broadcast"))
                processBroadcast(atts);
            break;

        case GetListing:
            if (level() != 3) {
                log_debug("Unknown XML tag in GetListing: %s", name);
                break;
            }

            if (!strcmp(name, "dir")) {
                const char *path;
                if (Attribute(atts).getString("path", path)
                    and !root->find(path)) {
                    DirNode *node = new DirNode;
                    root->insert(std::unique_ptr<DirNode>(node), path);
                    log_debug("Create dir %s", node->path().c_str());
                }
            }
            else if (!strcmp(name, "channel"))
                getChannel(atts);
            else if (!strcmp(name, "parameter"))
                getParameter(atts);
            break;

        case ReadData:
            if (level() != 3) {
                log_debug("Unknown XML tag in ReadData: %s", name);
                break;
            }

            if (name[0] == 'F' and periodic_subscriptions_) {
                Attribute a(atts);
                unsigned int index = 0;
                if (a.getUInt("c", index) and channel[index]) {
                    periodic_subscriptions_.readData(
                            *channel[index], a.find("d"));
                    // current_data_group_->readFromString(index, a.find("d"));
                }
            }
            else if (!strcmp(name, "time") and periodic_subscriptions_) {
                periodic_subscriptions_.readTimestamps(
                        Attribute(atts).find("d"));
            }

            break;

        case ReadEvent:
            if (level() != 3) {
                log_debug("Unknown XML tag in ReadEvent: %s", name);
                break;
            }

            if (name[0] == 'E') {
                Attribute a(atts);
                unsigned int index;
                if (a.getUInt("c", index)) {
                    const auto map_it = channel_event_.find(index);
                    if (map_it != channel_event_.end() and channel[index]) {
                        auto cache = variable_cache_.lock();
                        cache[0].readFromBase64(a.find("d"), *channel[index]);
                        for (const auto &weak_sub : map_it->second) {
                            if (const auto sub = weak_sub.lock()) {
                                sub->readData(cache[0].data());
                                sub->newValues(dataTime);
                            }
                        }
                    }
                }
            }

            break;

        default:
            break;
    };
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::endElement(const char *name)
{ /*{{{*/
    switch (state) {
        case GetListing:
            if (level() == 2) {
                state = Idle;

                // name is either
                //      - channels
                //      - parameters
                //      - listing
                if (!strcmp(name, "listing"))
                    processListRequest();
            }

            break;

        case ReadData:
            if (level() != 2 or strcmp(name, "data"))
                break;
            if (periodic_subscriptions_) {
                periodic_subscriptions_.endNewDataRecieving();
            }
            state = Idle;
            break;

        case ReadEvent:
            if (level() == 2 and !strcmp(name, "data")) {
                state = Idle;
            }
            break;

        default:
            break;
    }
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
// Arguments:
//     var: if var == 0, this is a test to check whether path was found
//          if var != 0, found the variable --> subscribe it
void ProtocolHandler::pendingAck(Variable *var)
{ /*{{{*/
    if (pending_queue_.empty())
        throw ProtocolError("Empty pending request queue");

    for (const auto &weak_sub : pending_queue_.front().subscriptions_) {
        auto ps = weak_sub.lock();
        if (!ps)
            continue;

        std::shared_ptr<PdCom::impl::Subscription> new_sub;
        if (var) {
            new_sub = subscribe(
                    ps->This_,
                    std::static_pointer_cast<const MsrProto::Variable>(
                            var->shared_from_this()),
                    ps->subscriber_, ps->getSelector(), false);
        }
        if (!new_sub) {
            new_sub = std::make_shared<InvalidSubscription>(
                    ps->This_, ps->subscriber_, process->shared_from_this());
            process->pending_callbacks_.push_back(
                    {ps, PdCom::Subscription::State::Invalid});
        }
        process->replace(ps, new_sub);
        ps->replaceImpl(std::move(new_sub));
    }

    if (!var) {
        // End of processing; clean up structures
        pending_queue_.pop_front();
    }
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::xsadAck()
{ /*{{{*/
    if (xsadQ.empty())
        throw ProtocolError("no pending xsad commands available");
    const auto ids = xsadQ.front();
    xsadQ.pop();
    if (ids.group_id == 0) {
        const auto map_it = channel_event_.find(ids.channel_id);
        if (map_it == channel_event_.end())
            return;
        for (const auto &weak_sub : map_it->second)
            if (const auto sub = weak_sub.lock())
                process->pending_callbacks_.push_back(
                        {sub, PdCom::Subscription::State::Active});
    }
    else {
        auto map_it =
                periodic_subscriptions_.at(ids.group_id).find(ids.channel_id);
        if (map_it == periodic_subscriptions_.at(ids.group_id).end())
            return;
        for (const auto &weak_sub : map_it->second.subscriptions_)
            if (const auto sub = weak_sub.lock())
                process->pending_callbacks_.push_back(
                        {sub, PdCom::Subscription::State::Active});
    }

} /*}}}*/

///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::xsapAck()
{ /*{{{*/
    if (xsapQ.empty())
        throw ProtocolError("no pending xsad commands available");

    auto map_it = parameter_event_.find(xsapQ.front());
    if (map_it != parameter_event_.end()) {
        for (const auto &weak_sub : map_it->second)
            if (const auto sub = weak_sub.lock())
                process->pending_callbacks_.push_back(
                        {sub, PdCom::Subscription::State::Active});
    }

    xsapQ.pop();
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
Parameter *ProtocolHandler::getParameter(const char **_atts)
{ /*{{{*/
    Attribute atts(_atts);
    unsigned index;

    if (!atts.getUInt("index", index))
        return 0;

    Parameter *&param = parameter[index];
    if (!param) {
        // index="0" name="/Constant/Value" datasize="8" typ="TDBL_LIST"
        // anz="10" cnum="10" rnum="1" orientation="VECTOR" flags="7"
        const char *name = atts.find("name");

        // Check index and name
        if (!*name) {
            log_debug("Missing index or name for variable. Ignoring");
            return 0;
        }

        // Check whether variable exists
        if (std::dynamic_pointer_cast<Variable>(root->find(name))) {
            log_debug("%s exists. Ignoring", name);
            return 0;
        }

        SizeInfo size_info;
        const auto type_info = getDataType(_atts, size_info);
        if (!type_info) {
            log_debug("Datatype for %s unspecified. Ignoring", name);
            return 0;
        }

        unsigned int flags;
        if (!atts.getUInt("flags", flags)) {
            log_debug("Flags for %s unspecified. Ignoring", name);
            return 0;
        }

        auto shared_param = std::make_shared<Parameter>(
                std::move(size_info), type_info, process->shared_from_this(),
                index, flags, atts.find("alias"), atts.isTrue("dir"));
        param = shared_param.get();
        root->insert(std::move(shared_param), name);
    }

    return param;
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
Channel *ProtocolHandler::getChannel(const char **_atts)
{ /*{{{*/
    Attribute atts(_atts);
    unsigned index;

    if (!atts.getUInt("index", index))
        return 0;

    Channel *&chan = channel[index];
    if (!chan) {
        // index="0" name="/Constant/Value" datasize="8" typ="TDBL_LIST"
        // anz="10" cnum="10" rnum="1" orientation="VECTOR" flags="7"
        const char *name = atts.find("name");

        // Check index and name
        if (!*name) {
            log_debug("Missing index or name for variable. Ignoring");
            return 0;
        }

        // Check whether variable exists
        if (std::dynamic_pointer_cast<Variable>(root->find(name))) {
            log_debug("%s exists. Ignoring", name);
            return 0;
        }

        SizeInfo size_info;
        const auto type_info = getDataType(_atts, size_info);
        if (!type_info) {
            log_debug("Datatype for %s unspecified. Ignoring", name);
            return 0;
        }

        unsigned int bufsize;
        if (!atts.getUInt("bufsize", bufsize)) {
            log_debug("bufsize for %s unspecified. Ignoring", name);
            return 0;
        }

        double freq;
        if (!atts.getDouble("HZ", freq) or !freq) {
            log_debug("HZ for %s unspecified or zero. Ignoring", name);
            return 0;
        }

        unsigned int taskId = 0;
        atts.getUInt("task", taskId);

        auto shared_chan = std::make_shared<Channel>(
                std::move(size_info), type_info, process->shared_from_this(),
                index, taskId, 1.0 / freq, atts.find("alias"), bufsize,
                atts.isTrue("dir"));
        chan = shared_chan.get();
        root->insert(std::move(shared_chan), name);
    }

    return chan;
} /*}}}*/

namespace {
struct TypeMap
{
    const char *msr_name;
    const PdCom::TypeInfo *type_info;
    size_t len;
    template <size_t N>
    constexpr TypeMap(const char (&name)[N], PdCom::TypeInfo const &ti) :
        msr_name(&name[0]), type_info(&ti), len(N - 1)
    {}
};
}  // namespace

///////////////////////////////////////////////////////////////////////////
const PdCom::TypeInfo *
ProtocolHandler::getDataType(const char **atts, PdCom::SizeInfo &size_info)
{ /*{{{*/
    unsigned int rnum       = 0;
    unsigned int cnum       = 0;
    unsigned int count      = 0;
    const char *orientation = 0;
    using PdCom::SizeInfo;
    using PdCom::TypeInfo;
    const TypeInfo *ans = nullptr;

    static constexpr TypeMap type_map[] = {
            {"TDBL", details::TypeInfoTraits<double>::type_info},
            {"TINT", details::TypeInfoTraits<int32_t>::type_info},
            {"TUINT", details::TypeInfoTraits<uint32_t>::type_info},
            {"TCHAR", details::TypeInfoTraits<int8_t>::type_info},
            {"TUCHAR", details::TypeInfoTraits<uint8_t>::type_info},
            {"TSHORT", details::TypeInfoTraits<int16_t>::type_info},
            {"TUSHORT", details::TypeInfoTraits<uint16_t>::type_info},
            {"TLINT", details::TypeInfoTraits<int64_t>::type_info},
            {"TULINT", details::TypeInfoTraits<uint64_t>::type_info},
            {"TFLT", details::TypeInfoTraits<float>::type_info},
    };

    const char *const typeStr = Attribute(atts).find("typ");

    // Try and find the name in msr_dtypemap
    for (const auto &t : type_map) {
        if (!strncmp(typeStr, t.msr_name, t.len))
            ans = t.type_info;
    }
    if (!ans)
        return nullptr;

    // cache variable data
    for (int i = 0; atts[i]; i += 2) {
        if (!strcmp(atts[i], "rnum")) {
            rnum = atoi(atts[i + 1]);
        }
        else if (!strcmp(atts[i], "cnum")) {
            cnum = atoi(atts[i + 1]);
        }
        else if (!strcmp(atts[i], "anz")) {
            count = atoi(atts[i + 1]);
        }
        else if (!strcmp(atts[i], "orientation")) {
            orientation = atts[i + 1];
        }
    }

    if (orientation) {
        if (!strcmp(orientation, "VECTOR")) {
            if (!count)
                return nullptr;

            size_info = SizeInfo::RowVector(count);
        }
        else if (orientation and !strncmp(orientation, "MATRIX", 6)) {
            if (!(rnum and cnum))
                return nullptr;

            size_info = SizeInfo::Matrix(rnum, cnum);
        }
        else
            return nullptr;
    }
    else {
        if (rnum > 1 or cnum > 1 or count > 1)
            return nullptr;

        size_info = SizeInfo();
    }

    return ans;
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
bool ProtocolHandler::login(const char *mech, const char *clientData)
{ /*{{{*/
    if (!feature.login)
        return false;

    XmlCommand sasl(cout, "auth");
    if (mech and *mech)
        sasl.addAttribute("mech", mech);

    if (clientData and *clientData)
        sasl.addAttribute("clientdata", clientData);

    return true;
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::logout()
{ /*{{{*/
    XmlCommand(cout, "auth").addAttribute("logout", 1);
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
bool ProtocolHandler::processLogin(const char **_atts)
{ /*{{{*/
    // pdserv sends an empty saslauth tag after logout()
    if (!*_atts)
        return false;
    Attribute atts(_atts);
    int finished = 0;
    if (atts.getInt("success", finished) and !finished)
        finished = -1;

    const char *mechlist = 0;
    atts.getString("mechlist", mechlist);

    const char *serverdata = 0;
    atts.getString("serverdata", serverdata);

    process->loginReply(mechlist, serverdata, finished);
    return finished > 0;
} /*}}}*/


///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::setPolite(bool state)
{ /*{{{*/
    if (state != polite and feature.polite) {
        polite = state;

        XmlCommand(cout, "remote_host").addAttribute("polite", state);
    }
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
bool ProtocolHandler::asyncData()
{ /*{{{*/
    std::unique_lock<AsyncDataMutex> lck(async_data_mutex_);
    char buf[1024];

    const auto n = StreambufLayer::read(buf, sizeof(buf));
    if (!protocolError and n > 0)
        parse(buf, n);
    return n != 0;
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
bool ProtocolHandler::find(const std::string &path)
{ /*{{{*/
    auto var = std::dynamic_pointer_cast<Variable>(root->find(path));
    if (var or path.empty()) {
        process->findReply(PdCom::impl::Variable::toUApi(var));
        return true;
    }

    findQ.push(path);

    if (fullListing == Uncached) {
        // Do not flush next command
        XmlCommand(cout, "rp").addEscapedAttribute("name", path).noFlush();

        XmlCommand(cout, "rk").addEscapedAttribute("name", path).setId("findQ");
    }

    return false;
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
bool ProtocolHandler::processFindRequest()
{                       /*{{{*/
    if (findQ.empty())  // required by listReply!
        return false;

    std::shared_ptr<DirNode> node = root->find(findQ.front());
    process->findReply(PdCom::impl::Variable::toUApi(
            std::dynamic_pointer_cast<PdCom::impl::Variable>(node)));
    findQ.pop();

    return !findQ.empty();
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
bool ProtocolHandler::list(const std::string &dir)
{ /*{{{*/
    listQ.push(dir);

    if (fullListing != Uncached) {
        if (fullListing == Cached)
            processListRequest();
    }
    else if (dir.empty() or !feature.list) {
        // List everything. First send a <rp> and the state machine
        // will send a <rp> thereafter
        XmlCommand(cout, "rk").noFlush();
        XmlCommand(cout, "rp").setId("listQ");

        fullListing = InProgress;
    }
    else
        XmlCommand(cout, "list")
                .addEscapedAttribute("path", dir)
                .addAttribute("noderived", 1);

    return fullListing == Cached;
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
bool ProtocolHandler::processListRequest()
{ /*{{{*/
    std::string path = listQ.front();
    listQ.pop();

    std::vector<std::string> dirList;

    DirNode::List list;
    const bool listall = path.empty();
    if (listall)
        root->getAllChildren(list);
    else {
        // Directory list
        std::shared_ptr<DirNode> node = root->find(path);
        if (node)
            node->getChildren(list);
    }

    std::vector<PdCom::Variable> varList;
    varList.reserve(list.size());

    for (const auto &node : list) {
        if (!listall and node->hasChildren())
            dirList.push_back(node->path());

        auto v = std::dynamic_pointer_cast<PdCom::impl::Variable>(node);
        if (v)
            varList.emplace_back(impl::Variable::toUApi(v));
    }

    process->listReply(varList, dirList);

    return !listQ.empty();
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::pollChannel(PollSubscription &sub)
{ /*{{{*/
    const auto index = static_cast<const Variable &>(*sub.variable_).index_;
    if (pending_channel_polls_.append(index, sub.shared_from_this()))
        XmlCommand(cout, "rk")
                .addAttribute("index", index)
                .addAttribute("short", 1)
                .setId("pollChannel");
} /*}}}*/

void ProtocolHandler::pollParameter(PollSubscription &sub)
{
    const auto index = static_cast<const Variable &>(*sub.variable_).index_;
    if (pending_parameter_polls_.append(index, sub.shared_from_this()))
        pollParameter(index, "pollParameter");
}

///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::pollParameter(size_t index, const char *id)
{ /*{{{*/
    XmlCommand(cout, "rp")
            .addAttribute("index", index)
            .addAttribute("short", 1)
            .addAttribute("hex", 1)
            .setId(id);
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
std::shared_ptr<PdCom::impl::Subscription> ProtocolHandler::subscribe(
        PdCom::Subscription *subscription,
        std::shared_ptr<const PdCom::impl::Variable> _var,
        PdCom::Subscriber &subscriber,
        const PdCom::Selector &selector,
        bool notify_pending)
{
    auto var = std::dynamic_pointer_cast<const Variable>(_var);
    if (!var)
        return nullptr;
    if (subscriber.getTransmission() == PdCom::poll_mode) {
        auto ans = std::make_shared<PollSubscription>(
                var, subscription, subscriber, process->shared_from_this(),
                selector);
        process->pending_callbacks_.push_back(
                {ans, PdCom::Subscription::State::Active});
        return ans;
    }
    else if (subscriber.getTransmission() == PdCom::event_mode) {
        auto &sub_map = var->is_parameter_ ? parameter_event_ : channel_event_;
        auto ans      = std::make_shared<EventSubscription>(
                var, subscription, subscriber, process->shared_from_this(),
                selector);
        if (sub_map[var->index_].empty()) {
            if (var->is_parameter_ and feature.xsap) {
                XmlCommand(cout, "xsap")
                        .addAttribute("parameters", var->index_)
                        .addAttribute("hex", 1)
                        .setId("xsapQ");
                xsapQ.push(var->index_);
            }
            else if (!var->is_parameter_) {
                XmlCommand(cout, "xsad")
                        .addAttribute("channels", var->index_)
                        .addAttribute("group", 0)
                        .addAttribute("coding", "Base64")
                        .addAttribute("event", 1)
                        .setId("xsadQ");
                xsadQ.push({0, var->index_});
            }
            if (var->is_parameter_ and !feature.xsap) {
                process->pending_callbacks_.push_back(
                        {ans, PdCom::Subscription::State::Active});
            }
            else if (notify_pending)
                process->pending_callbacks_.push_back(
                        {ans, PdCom::Subscription::State::Pending});
        }
        else {
            for (const auto &weak_s : sub_map[var->index_]) {
                if (const auto s = weak_s.lock()) {
                    if (!(s->This_->getState()
                                  == PdCom::Subscription::State::Pending
                          and !notify_pending)) {
                        process->pending_callbacks_.push_back(
                                {ans, s->This_->getState()});
                        break;
                    }
                }
            }
        }
        sub_map[var->index_].insert(ans);
        return ans;
    }
    else {
        if (var->is_parameter_)
            return {};
        auto ans = std::make_shared<PeriodicSubscription>(
                var, subscription, subscriber, process->shared_from_this(),
                selector);
        const auto si = periodic_subscriptions_.add(
                process->pending_callbacks_, notify_pending, ans);
        if (si.needs_server_action) {
            XmlCommand(cout, "xsad")
                    .addAttribute("channels", var->index_)
                    .addAttribute("group", si.group_id)
                    .addAttribute("coding", "Base64")
                    .addAttribute("reduction", si.decimation_)
                    .addAttribute("blocksize", si.blocksize_)
                    .setId("xsadQ");
            xsadQ.push({si.group_id, var->index_});
        }
        return ans;
    }
}


///////////////////////////////////////////////////////////////////////////
std::shared_ptr<PdCom::impl::Subscription> ProtocolHandler::subscribe(
        PdCom::Subscription *subscription,
        const std::string &path,
        PdCom::Subscriber &subscriber,
        const PdCom::Selector &selector)
{ /*{{{*/
    setPolite(false);

    if (auto var = std::dynamic_pointer_cast<Variable>(root->find(path))) {
        return subscribe(subscription, var, subscriber, selector, false);
    }

    // Variable is unknown. Find it first. This requires interaction
    // with the server, which automatically postpones calls to
    // request->subscriber to a later point in time
    const auto it = std::find_if(
            pending_queue_.begin(), pending_queue_.end(),
            [&path](const PendingRequest &pr) { return pr.path_ == path; });

    auto ans = std::make_shared<PendingSubscription>(
            subscription, subscriber, process->shared_from_this(),
            selector.impl_);

    if (it == pending_queue_.end()) {
        // Discovery not yet active
        XmlCommand(cout, "rp")
                .addAttribute("name", path)
                .addAttribute("hex", 1)
                .setId("pendingParameter")
                .noFlush();
        XmlCommand(cout, "rk")
                .addAttribute("name", path)
                .setId("pendingChannel");
        pending_queue_.push_back({path, {{ans}}});
    }
    else {
        it->subscriptions_.insert(ans);
    }
    process->pending_callbacks_.push_back(
            {ans, PdCom::Subscription::State::Pending});
    return ans;
} /*}}}*/

void ProtocolHandler::unsubscribe(EventSubscription &s) noexcept
{
    const PthreadCancelDisabler di;
    const auto &var   = static_cast<Variable const &>(*s.variable_);
    auto &sub_map     = var.is_parameter_ ? parameter_event_ : channel_event_;
    const auto map_it = sub_map.find(var.index_);
    if (map_it != sub_map.end()) {
        if (map_it->second.remove_expired()) {
            try {
                if (var.is_parameter_ and feature.xsap) {
                    XmlCommand(cout, "xsop")
                            .addAttribute("parameters", var.index_);
                }
                else {
                    XmlCommand(cout, "xsod")
                            .addAttribute("channels", var.index_)
                            .addAttribute("group", 0);
                }
            }
            catch (...) {
            }
        }
    }
}
void ProtocolHandler::unsubscribe(PeriodicSubscription &s) noexcept
{
    const PthreadCancelDisabler di;
    const auto si = periodic_subscriptions_.remove(s);
    if (si.needs_server_action) {
        try {
            XmlCommand(cout, "xsod")
                    .addAttribute(
                            "channels",
                            static_cast<const Variable &>(*s.variable_).index_)
                    .addAttribute("group", si.group_id);
        }
        catch (...) {
        }
    }
}

///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::ping()
{ /*{{{*/
    XmlCommand(cout, "ping");
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
std::string ProtocolHandler::name() const
{ /*{{{*/
    return m_name;
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
std::string ProtocolHandler::version() const
{ /*{{{*/
    return m_version;
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
void ProtocolHandler::writeParameter(
        const Parameter &p,
        const void *src,
        TypeInfo::DataType src_type,
        size_t idx,
        size_t n)
{ /*{{{*/
    XmlCommand(cout, "wp")
            .addAttribute("index", p.index_)
            .addAttribute("startindex", idx, idx)
            .addAttributeStream(
                    "hexvalue",
                    [&, this](std::ostream &os) {
                        encoder_.toHex(*p.type_info, src, src_type, n, os);
                    })
            .setId("wp");
} /*}}}*/

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
bool fcmp(const char *p, const char *target, size_t len)
{
    return len >= ::strlen(target) && !strncmp(p, target, len);
}

void ProtocolHandler::Feature::operator=(const char *str)
{ /*{{{*/
    if (!str)
        return;
    *this = Feature {};

    const char *p = str;
    size_t len;

    do {
        p   = str;
        str = strchr(str, ',');

        len = str ? str++ - p : strlen(p);

        if (len) {
            if (fcmp(p, "pushparameters", len))
                pushparameters = true;

            if (fcmp(p, "binparameters", len))
                binparameters = true;

            if (fcmp(p, "eventchannels", len))
                eventchannels = true;

            if (fcmp(p, "statistics", len))
                statistics = true;

            if (fcmp(p, "pmtime", len))
                pmtime = true;

            if (fcmp(p, "aic", len))
                aic = true;

            if (fcmp(p, "messages", len))
                messages = true;

            if (fcmp(p, "quiet", len))
                quiet = true;

            if (fcmp(p, "list", len))
                list = true;

            if (fcmp(p, "exclusive", len))
                exclusive = true;

            if (fcmp(p, "polite", len))
                polite = true;

            if (fcmp(p, "xsadgroups", len))
                xsadgroups = true;

            if (fcmp(p, "xsap", len))
                xsap = true;

            if (fcmp(p, "login", 5))
                login = true;

            if (fcmp(p, "eventid=", 8))
                eventid = strtoul(p + 8, 0, 0);
        }

    } while (str);
} /*}}}*/


void ProtocolHandler::broadcast(
        const std::string &message,
        const std::string &attr)
{
    XmlCommand msg(cout, "broadcast");
    msg.addEscapedAttribute(attr.c_str(), message);
}


void ProtocolHandler::processBroadcast(const char **atts)
{
    Attribute a(atts);
    const char *try_keys[] = {"action", "text"};
    const char *value = nullptr, *user = nullptr;
    if (!a.getString("user", user))
        user = "anonymous";
    for (const char *key : try_keys) {
        if (a.getString(key, value)) {
            process->broadcastReply(value, key, a.getTime_ns("time"), user);
            return;
        }
    }
}

template class PdCom::impl::MsrProto::ExpatWrapper<ProtocolHandler>;

bool ProtocolHandler::PendingPollList::append(
        unsigned int const id,
        const std::shared_ptr<PollSubscription> &sub)
{
    for (auto &pp : *this) {
        if (pp.id_ == id) {
            pp.subscriptions_.insert(sub);
            pp.subscriptions_.remove_expired();
            return false;
        }
    }
    push_back({id, {{sub}}});
    return true;
}
